/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "dump_gui.h"

#include <gtk/gtk.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "error.h"
#include "mark_gui.h"
#include "version.h"
#include "audio.h"
#include "mixer.h"
#include "yes_no_gui.h"
#include "file_gui.h"
#include "endian.h"
#include "wav.h"

static int gui_on = 0;

static GtkWidget *win;
static GtkWidget *combo1, *combo2;
static GtkWidget *combo3;
static GtkWidget *filename_entry;

/* dump types */
#define DUMP_WAV_STEREO		0
#define DUMP_WAV_MONO		1
#define DUMP_FLOAT_STEREO	2
#define DUMP_FLOAT_MONO		3

static int ready;
static char filename[FILENAME_MAX+1];
static int start_point, end_point;
static int type;
static int user_interrupt;

static void destroy_event(void)
{
  user_interrupt = 1;
}

static GtkWidget *bar;

static void calculate(GtkWidget *win)
{
  FILE *f;
  static char buf[4096];
  int nb = end_point - start_point + 1;
  int i;
  int max_size = 0;
  int mul_factor = 0;
  int (*fn)(mixer *, void *, int) = 0;

  switch (type) {
    case DUMP_WAV_STEREO:
      max_size = 1024;
      mul_factor = 4;
      fn = main_mixer->mix;
      break;

    case DUMP_WAV_MONO:
      max_size = 2048;
      mul_factor = 2;
      fn = main_mixer->mix_mono;
      break;

    case DUMP_FLOAT_STEREO:
      max_size = 512;
      mul_factor = 8;
      fn = main_mixer->float_mix;
      break;

    case DUMP_FLOAT_MONO:
      max_size = 1024;
      mul_factor = 4;
      fn = main_mixer->float_mono_mix;
      break;
  }

  f = fopen(filename, "w");
  if (!f) {
    int serr=errno;
    gtk_widget_destroy(win);
    reset_error();
    add_error_char(filename);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return;
  }

  if (type == DUMP_WAV_STEREO || type == DUMP_WAV_MONO) {
    wav_stereo_create_header(buf, end_point-start_point+1, type == DUMP_WAV_STEREO);

    if (fwrite(buf, 44, 1, f) != 1) {
      int serr=errno;
      gtk_widget_destroy(win);
      fclose(f);
      unlink(filename);
      reset_error();
      add_error_char(filename);
      add_error_char(": ");
      add_error_char(strerror(serr));
      add_error_char("\nHint: you may have a disk full.\n");
      display_error();
      return;
    }
  }

  main_mixer->has_last_sample = 1;
  main_mixer->next_sample = start_point;
  main_mixer->last_sample = end_point;

  main_mixer->reset_effects(main_mixer);

more:
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(bar), (float)(end_point-start_point+1-nb)/(float)(end_point-start_point+1));

  while (gtk_events_pending())
    gtk_main_iteration();

  if (user_interrupt) {
    fclose(f);
    unlink(filename);
    return;
  }

  for (i=0; i<10; i++) {
    int size = nb;
    if (size == 0)
      goto done;
    if (size > max_size)
      size = max_size;
    if (fn(main_mixer, buf, size) == -1)
      goto done;
    if (fwrite(buf, size*mul_factor, 1, f) != 1) {
      int serr = errno;
      gtk_widget_destroy(win);
      fclose(f);
      unlink(filename);
      reset_error();
      add_error_char(filename);
      add_error_char(": ");
      add_error_char(strerror(serr));
      add_error_char("\nHint: you may have a disk full.\n");
      display_error();
      return;
    }
    nb -= size;
  }

  goto more;

done:
  gtk_widget_destroy(win);

  if (fclose(f) == -1) {
    int serr=errno;
    unlink(filename);
    reset_error();
    add_error_char(filename);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return;
  }
}

static void do_dump(void)
{
  extern GtkWidget *main_win;
  GtkWidget *win;
  GtkWidget *box;
  GtkWidget *hbox;
  GtkWidget *cancel;

  user_interrupt = 0;

  /* gui */
  win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(win), 5);

  box = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(win), box);

  gtk_box_pack_start(GTK_BOX(box), gtk_label_new("Dumping to:"), FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(box), gtk_label_new(filename), FALSE, FALSE, 0);

  bar = gtk_progress_bar_new();
  gtk_box_pack_start(GTK_BOX(box), bar, FALSE, FALSE, 0);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(bar), 0);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
  cancel = gtk_button_new_with_mnemonic("_Cancel");
  gtk_box_pack_end(GTK_BOX(hbox), cancel, FALSE, FALSE, 0);

  /* signals */
  g_signal_connect(win, "destroy", G_CALLBACK(destroy_event), 0);
  g_signal_connect_swapped(cancel, "clicked", G_CALLBACK(gtk_widget_destroy), win);

  gtk_window_set_modal(GTK_WINDOW(win), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(win), GTK_WINDOW(main_win));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);

  gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER_ON_PARENT);

  /* done */
  gtk_widget_show_all(win);

  calculate(win);
}

static void go_event(void)
{
  char *c = (char *)gtk_entry_get_text(GTK_ENTRY(filename_entry));
  mark *t;
  struct stat buf;

  if (strlen(c) == 0) {
    reset_error();
    add_error_char("You must provide a file name\n");
    display_error_modal();
    return;
  }
  if (strlen(c) > FILENAME_MAX || strlen(c) < 0) {
    reset_error();
    add_error_char(c);
    add_error_char("This file name is too long (");
    add_error_int(FILENAME_MAX);
    add_error_char(" bytes is the maximum).\n");
    display_error_modal();
    return;
  }

  if (stat(c, &buf) != -1) {
    /* error used as buffer */
    reset_error();
    add_error_char(c);
    add_error_char("\nThe file already exist, do you want to overwrite it ?\n");
    if (!yes_no(win, get_error(), 0, 0))
      return;
  }

  strcpy(filename, c);  /* no overflow possible */

  t = get_mark((char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(combo1)->entry)));
  if (!t)
    start_point = first_audio_sample();
  else
    start_point = t->sample;

  t = get_mark((char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(combo2)->entry)));
  if (!t)
    end_point = last_audio_sample();
  else
    end_point = t->sample;

  if (end_point < start_point) {
    reset_error();
    add_error_char("The starting point comes after the ending point.\n");
    display_error_modal();
    return;
  }

  ready = 1;

  c = (char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(combo3)->entry));
  if (!strcmp(c, "44.1 KHz stereo WAV"))
    type = DUMP_WAV_STEREO;
  else if (!strcmp(c, "44.1 KHz mono WAV"))
    type = DUMP_WAV_MONO;
  else if (!strcmp(c, "44.1 KHz stereo float 32 bits"))
    type = DUMP_FLOAT_STEREO;
  else
    type = DUMP_FLOAT_MONO;

  gtk_widget_destroy(win);
}

static void set_filename(char *c)
{
  gtk_entry_set_text(GTK_ENTRY(filename_entry), c);
  gtk_editable_set_position(GTK_EDITABLE(filename_entry), -1);
}

static void choose_file_event(void)
{
  new_file_selector("Choose a file where to dump the audio mix", set_filename, win);
}

void dump_mix(void)
{
  extern GtkWidget *main_win;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *choose_filename;
  GtkWidget *ok, *cancel;
  GList *list1 = 0, *list2 = 0;
  GList *list3 = 0;
  GList *new;
  int i;

  if (gui_on)
    return;

  if (big_endian) {
    reset_error();
    add_error_char("The dump tool does not work for big endian machines. Request the authors to do it.\n");
    display_error();
    return;
  }

  /* audio might be on */
  close_audio();

  /* data */
  new = g_list_append(list1, "First audio sample"); if (!new) goto data_error; list1 = new;
  for (i=0; i<mark_list_size; i++) {
    new = g_list_append(list1, mark_list[i]->name); if (!new) goto data_error; list1 = new;
  }

  new = g_list_append(list2, "Last audio sample"); if (!new) goto data_error; list2 = new;
  for (i=0; i<mark_list_size; i++) {
    new = g_list_append(list2, mark_list[i]->name); if (!new) goto data_error; list2 = new;
  }

  new = g_list_append(list3, "44.1 KHz stereo WAV"); if (!new) goto data_error; list3 = new;
  new = g_list_append(list3, "44.1 KHz mono WAV"); if (!new) goto data_error; list3 = new;
  new = g_list_append(list3, "44.1 KHz stereo float 32 bits"); if (!new) goto data_error; list3 = new;
  new = g_list_append(list3, "44.1 KHz mono float 32 bits"); if (!new) goto data_error; list3 = new;

  /* gui */
  win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(win), 5);
  gtk_window_set_transient_for(GTK_WINDOW(win), GTK_WINDOW(main_win));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);
  gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER_ON_PARENT);
  /* error used as buffer */
  reset_error();
  add_error_char("Songs ");
  add_error_char(VERSION);
  add_error_char(" - dump to file");
  gtk_window_set_title(GTK_WINDOW(win), get_error());
  gtk_grab_add(win);

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(win), vbox);

  /* first line - choose file name */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Dump to"), FALSE, FALSE, 0);
  filename_entry = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox), filename_entry, TRUE, TRUE, 0);
  choose_filename = gtk_button_new_with_label("Choose");
  gtk_box_pack_start(GTK_BOX(hbox), choose_filename, FALSE, FALSE, 0);

  /* second line - choose start/end pos */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("From"), FALSE, FALSE, 0);
  combo1 = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(hbox), combo1, TRUE, TRUE, 0);
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo1)->entry), FALSE);
  gtk_combo_set_popdown_strings(GTK_COMBO(combo1), list1);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("to"), FALSE, FALSE, 0);
  combo2 = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(hbox), combo2, TRUE, TRUE, 0);
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo2)->entry), FALSE);
  gtk_combo_set_popdown_strings(GTK_COMBO(combo2), list2);

  /* third line - file type */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("File type"), FALSE, FALSE, 0);
  combo3 = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(hbox), combo3, TRUE, TRUE, 0);
  gtk_entry_set_editable(GTK_ENTRY(GTK_COMBO(combo3)->entry), FALSE);
  gtk_combo_set_popdown_strings(GTK_COMBO(combo3), list3);

  /* fourth line - OK Cancel */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  ok = gtk_button_new_with_mnemonic("_OK");
  cancel = gtk_button_new_with_mnemonic("_Cancel");
  gtk_box_pack_end(GTK_BOX(hbox), ok, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(hbox), cancel, FALSE, FALSE, 0);

  /* callbacks */
  g_signal_connect(choose_filename, "clicked", G_CALLBACK(choose_file_event), 0);
  g_signal_connect(ok, "clicked", G_CALLBACK(go_event), 0);
  g_signal_connect_swapped(cancel, "clicked", G_CALLBACK(gtk_widget_destroy), win);
  g_signal_connect(win, "destroy", G_CALLBACK(gtk_main_quit), 0);

  gtk_widget_show_all(win);

  gui_on = 1;

  ready = 0;

  gtk_main();

  gui_on = 0;

  if (!ready)
    return;

  do_dump();
  return;

data_error:
  if (list1) g_list_free(list1);
  if (list2) g_list_free(list2);
  if (list3) g_list_free(list3);
  reset_error();
  oom();
  display_error();
}
