/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "file.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include "track.h"
#include "mixer.h"
#include "error.h"
#include "channel.h"
#include "main_gui.h"
#include "version.h"
#include "input.h"
#include "pos.h"
#include "vol.h"
#include "pan.h"
#include "sc1.h"
#include "track_gui.h"
#include "effects_gui.h"
#include "mark_gui.h"
#include "version.h"

static char *current_file;

static char *buf = 0;
static int size = 0;
static int maxsize = 0;
static int un_putted = 0;

static void free_buffer(void)
{
  if (maxsize) {
    free(buf);
    buf = 0;
    maxsize = size = 0;
  }
}

static void reset_buffer(void)
{
  size = 0;
}

static int put_char(char c)
{
  if (size == maxsize) {
    char *new = realloc(buf, maxsize + 256);
    if (!new) {
      oom();
      display_error();
      return -1;
    }
    buf = new;
    maxsize += 256;
  }

  buf[size] = c;
  size++;

  return 0;
}

static void unget_line(void)
{
  un_putted = 1;
}

/* return 0 in case of EOF */
static int get_line(FILE *f)
{
  int c;

  if (un_putted) {
    un_putted = 0;
    return 0;
  }

again:
  reset_buffer();

  while(1) {
    c = getc(f);
    if (c == -1) {
      /* if nothing was read, EOF => void line must be returned */
      if (!size)
        return 0;
      /* otherwise, we have some data there, it means no \n to the end */
      break;
    }

    if (c == '\n')
      break;

    if (size == 0 && isspace(c))
      continue;

    if (put_char(c) == -1)
      return -1;
  }

  /* we catch void lines here */
  if (!size)
    goto again;

  /* comments here */
  if (size && buf[0] == '#')
    goto again;

  /* a line must have 0 at end */
  if (size)
    if (put_char(0) == -1)
      return -1;

  return 0;
}

static track **track_list = 0;
static int track_list_size = 0;
static int track_list_maxsize = 0;

static track *get_track(void)
{
  track *nt;
  if (track_list_size == track_list_maxsize) {
    track **n = realloc(track_list, (track_list_maxsize + 128) * sizeof(track *));
    if (!n) {
      oom();
      display_error();
      return 0;
    }
    track_list = n;
    track_list_maxsize+=128;
  }

  nt = malloc(sizeof(track));
  if (!nt) {
    oom();
    display_error();
    return 0;
  }

  track_list_size++;

  memset(nt, 0, sizeof(track));
  track_list[track_list_size-1] = nt;

  return nt;
}

static char **track_names = 0;
static int track_names_size = 0;
static int track_names_maxsize = 0;

static int add_name(char *c)
{
  if (track_names_size == track_names_maxsize) {
    char **n = realloc(track_names, (track_names_maxsize + 128) * sizeof(char *));
    if (!n) {
      oom();
      display_error();
      return -1;
    }
    track_names = n;
    track_names_maxsize += 128;
  }

  track_names[track_names_size] = strdup(c);

  if (!track_names[track_names_size]) {
    oom();
    display_error();
    return -1;
  }

  track_names_size++;

  return 0;
}

static mark **m_list = 0;
static int m_size = 0;
static int m_maxsize = 0;

mark *add_mark(void)
{
  mark *m;

  if (m_size == m_maxsize) {
    mark **n = realloc(m_list, (m_maxsize + 32) * sizeof(mark *));
    if (!n)
      return 0;
    m_list = n;
    m_maxsize += 32;
  }

  m = malloc(sizeof(mark));
  if (!m)
    return 0;

  memset(m, 0, sizeof(mark));

  /* marks are on by default */
  m->on = 1;

  m_list[m_size] = m;

  m_size++;

  return m;
}

/* return -1 if no mem, -2 if name already in use, 0 if ok */
int set_name(mark *m, char *name)
{
  int i;

  /* m_size-1 because the last in the list is us */
  for (i=0; i<m_size-1; i++)
    if (!strcmp(name, m_list[i]->name))
      return -2;

  m->name = strdup(name);
  if (!m->name)
    return -1;

  return 0;
}

/* return -1 if error, 0 if ok */
int set_position(mark *m, char *pos)
{
  int sample;
  int y_position;

  if (sscanf(pos, "%d %d", &sample, &y_position) != 2)
    return -1;

  m->sample = sample;
  m->y_position = y_position;

  return 0;
}

int load(char *name)
{
  FILE *f = fopen(name, "r");
  int major, minor;
  int save_modify = modify;
  int i;
  int nb;
  int ret;
  mark *marker;
  channel *mixer_channel = &main_mixer->_raw_mix;
  char *buf2;

  current_file = name;

  if (!f) {
    int serr = errno;
    reset_error();
    add_error_char("Load\n");
    add_error_char(name);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    goto error;
  }

  /* first line is identification line */
  if (get_line(f) == -1 || size == 0)
    goto error0;

  if (sscanf(buf, "[Songs %d.%d]", &major, &minor) != 2)
    goto error0;

  /* version verification */
  if (major > VERSION_MAJOR || minor > VERSION_MINOR) {
    reset_error();
    add_error_char("Load\n");
    add_error_char("The file ");
    add_error_char(current_file);
    add_error_char(" has version ");
    add_error_int(major);
    add_error_char(".");
    add_error_int(minor);
    add_error_char("\nwhich is not supported in this version (");
    add_error_char(VERSION);
    add_error_char(") of Songs\n");
    display_error();
    goto error;
  }

  /* [Marker] stuff */
more_marker:
  if (get_line(f) == -1 || size == 0)
    goto error0;

  if (strncmp("[Marker]", buf, strlen("[Marker]"))) {
    unget_line();
    goto marker_done;
  }

  marker = add_mark();
  if (!marker)
    goto error;

  if (strstr(buf, "(hide)"))
    marker->on = 0;

  if (get_line(f) == -1 || size == 0)
    goto error0;

  if (strncmp("Name", buf, strlen("Name")))
    goto error0;

  buf2 = buf+4;
  while(*buf2 && *buf2 != '=') {
    if (!isspace(*buf2)) goto error0;
    buf2++;
  }
  if (*buf2 == 0) goto error0;
  buf2++;
  while (isspace(*buf2)) buf2++;
  if (*buf2 == 0) goto error0;
  ret =  set_name(marker, buf2);
  if (ret == -1) goto error;
  if (ret == -2) goto error0;

  if (get_line(f) == -1 || size == 0)
    goto error0;

  if (strncmp("Position", buf, strlen("Position")))
    goto error0;

  buf2 = buf+8;
  while(*buf2 && *buf2 != '=') {
    if (!isspace(*buf2)) goto error0;
    buf2++;
  }
  if (*buf2 == 0) goto error0;
  buf2++;
  while (isspace(*buf2)) buf2++;
  if (*buf2 == 0) goto error0;
  if (set_position(marker, buf2) == -1) goto error0;

  goto more_marker;

marker_done:

  /* [Mixer] stuff */
  if (get_line(f) == -1 || size == 0)
    goto error0;

  if (strncmp("[Mixer]", buf, strlen("[Mixer]")))
    goto error0;

  /* we have a list of effects */
  while (1) {
    if (get_line(f) == -1 || size == 0)
      goto error0;

    if (!strncmp(buf, "Panning", strlen("Panning"))) {
      pan *p;
      double pan;
      if (sscanf(buf, "Panning = %lf", &pan) != 1)
        goto error0;
      p = new_pan(mixer_channel);
      if (!p)
        goto error;
      mixer_channel = (channel *)p;
      set_pan(p, pan);
      if (strstr(buf, "(bypass)"))
        p->c.bypass = 1;
      continue;
    }

    if (!strncmp(buf, "Gain", strlen("Gain"))) {
      vol *v;
      double vol;
      if (sscanf(buf, "Gain = %lf", &vol) != 1)
        goto error0;
      v = new_vol(mixer_channel);
      if (!v)
        goto error;
      mixer_channel = (channel *)v;
      set_vol(v, vol);
      if (strstr(buf, "(bypass)"))
        v->c.bypass = 1;
      continue;
    }

    if (!strncmp(buf, "Compressor", strlen("Compressor"))) {
      sc1 *s;
      double attack;
      double release;
      double threshold;
      double ratio;
      double knee;
      double makeup_gain;
      if (sscanf(buf, "Compressor = %lf %lf %lf %lf %lf %lf", &attack, &release, &threshold, &ratio, &knee, &makeup_gain) != 6)
        goto error0;
      s = new_sc1(mixer_channel);
      if (!s)
        goto error;
      mixer_channel = (channel *)s;
      set_attack(s, attack);
      set_release(s, release);
      set_threshold(s, threshold);
      set_ratio(s, ratio);
      set_knee(s, knee);
      set_makeup_gain(s, makeup_gain);
      if (strstr(buf, "(bypass)"))
        s->c.bypass = 1;
      continue;
    }

    /* not a valid effect, maybe the next track ? we unget the line */
    unget_line();
    break;
  }

  /* read the tracks */
  while (1) {
    track *n;

    if (get_line(f) == -1)
      goto error0;
    if (!size)
      break;

    /* so we have a new track to load, let's get some room to save it */
    n = get_track();
    if (!n)
      goto error;

    /* first line must be this */
    if (strncmp("[Track]", buf, strlen("[Track]")))
      goto error0;

    if (strstr(buf, "(mute)"))
      n->mute = 1;

    /* then the name of the track */
    if (get_line(f) == -1 || !size)
      goto error0;

    if (strncmp("Name", buf, strlen("Name")))
      goto error0;

    buf2 = buf + 4;
    while(*buf2 && *buf2 != '=') {
      if (!isspace(*buf2)) goto error0;
      buf2++;
    }
    if (*buf2 == 0) goto error0;
    buf2++;
    while (isspace(*buf2)) buf2++;
    if (*buf2 == 0) goto error0;

    if (add_name(buf2) == -1)
      goto error0;

    /* then we must have the file (it MUST be the first stuff defined for the track */
    if (get_line(f) == -1 || !size)
      goto error0;

    if (strncmp("File", buf, strlen("File")))
      goto error0;

    buf2 = buf+4;
    while(*buf2 && *buf2 != '=')	/* we may have garbage between "File" and "=", but we may have valid options, so... */
      buf2++;
    if (*buf2 == 0) goto error0;
    buf2++;
    while (isspace(*buf2)) buf2++;
    if (*buf2 == 0) goto error0;

    if (strstr(buf, "(wav mono)"))
      n->channel = (channel *)new_input(buf2, WAV_TYPE_MONO);
    else
    if (strstr(buf, "(wav stereo left)"))
      n->channel = (channel *)new_input(buf2, WAV_TYPE_STEREO_LEFT);
    else
    if (strstr(buf, "(wav stereo right)"))
      n->channel = (channel *)new_input(buf2, WAV_TYPE_STEREO_RIGHT);
    else
    if (strstr(buf, "(raw 16b little endian mono)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_16B_LITTLE_ENDIAN_MONO);
    else
    if (strstr(buf, "(raw 16b little endian stereo left)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT);
    else
    if (strstr(buf, "(raw 16b little endian stereo right)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT);
    else
    if (strstr(buf, "(raw float mono)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_FLOAT_MONO);
    else
    if (strstr(buf, "(raw float stereo left)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_FLOAT_STEREO_LEFT);
    else
    if (strstr(buf, "(raw float stereo right)"))
      n->channel = (channel *)new_input(buf2, RAW_TYPE_FLOAT_STEREO_RIGHT);
    else
      /* a dummy type */
      n->channel = (channel *)new_input(buf2, RAW_TYPE_16B_LITTLE_ENDIAN_MONO);
    if (!n->channel)
      goto error;

    n->input = (input *)n->channel;

    /* then Position stuff */
    if (get_line(f) == -1 || !size)
      goto error0;

    {
      pos *p;
      int global_start, subchannel_start, subchannel_nb_samples;
      if (sscanf(buf, "Position = %d %d %d", &global_start, &subchannel_start, &subchannel_nb_samples) != 3)
        goto error0;
      p = new_pos(n->channel);
      if (!p)
        goto error;
      n->channel = (channel *)p;
      set_global_start(p, global_start);
      set_subchannel_start(p, subchannel_start);
      set_subchannel_nb_samples(p, subchannel_nb_samples);
    }

    /* then, we have the effects, one after the other */
    while (1) {
      if (get_line(f) == -1)
        goto error0;
      if (!size)
        break;

      if (!strncmp(buf, "Panning", strlen("Panning"))) {
        pan *p;
        double pan;
        if (sscanf(buf, "Panning = %lf", &pan) != 1)
          goto error0;
        p = new_pan(n->channel);
        if (!p)
          goto error;
        n->channel = (channel *)p;
        set_pan(p, pan);
        if (strstr(buf, "(bypass)"))
          p->c.bypass = 1;
        continue;
      }

      if (!strncmp(buf, "Gain", strlen("Gain"))) {
        vol *v;
        double vol;
        if (sscanf(buf, "Gain = %lf", &vol) != 1)
          goto error0;
        v = new_vol(n->channel);
        if (!v)
          goto error;
        n->channel = (channel *)v;
        set_vol(v, vol);
        if (strstr(buf, "(bypass)"))
          v->c.bypass = 1;
        continue;
      }

      if (!strncmp(buf, "Compressor", strlen("Compressor"))) {
        sc1 *s;
        double attack;
        double release;
        double threshold;
        double ratio;
        double knee;
        double makeup_gain;
        if (sscanf(buf, "Compressor = %lf %lf %lf %lf %lf %lf", &attack, &release, &threshold, &ratio, &knee, &makeup_gain) != 6)
          goto error0;
        s = new_sc1(n->channel);
        if (!s)
          goto error;
        n->channel = (channel *)s;
        set_attack(s, attack);
        set_release(s, release);
        set_threshold(s, threshold);
        set_ratio(s, ratio);
        set_knee(s, knee);
        set_makeup_gain(s, makeup_gain);
        if (strstr(buf, "(bypass)"))
          s->c.bypass = 1;
        continue;
      }

      /* not a valid effect, maybe the next track ? we unget the line */
      unget_line();
      break;
    }
  }

  /* we are done ! */
  fclose(f);
  free_buffer();

  /* we must replace main_mixer mixer_channel */
  delete_channel_and_sub(main_mixer->mixer_track.channel);
  main_mixer->mixer_track.channel = mixer_channel;

  /* we must remove all tracks from the main_mixer */
  nb = main_mixer->nb_tracks;
  for (i = 0; i<nb; i++) {
    track *t = mixer_get_track(main_mixer, 0);
    delete_track(t);
  }

  /* and add the new ones */
  for (i=0; i<track_list_size; i++) {
    if (mixer_add_track(main_mixer, track_list[i]) == -1)
      goto error_add;
    track_gui(track_list[i]);
    main_gui_add_track(track_list[i]);
    set_track_name(track_list[i], track_names[i]);
  }
  /* we need to call this one only once */
  if (track_list_size)
    effects_gui_new_track(track_list[0]);

  if (track_list_maxsize) {
    free(track_list);
    track_list = 0;
    track_list_size = track_list_maxsize = 0;
  }

  for (i=0; i<track_names_size; i++)
    free(track_names[i]);
  if (track_names_maxsize) {
    free(track_names);
    track_names = 0;
    track_names_size = track_names_maxsize = 0;
  }

  /* we must remove all marks */
  nb = mark_list_size;
  for (i=0; i<nb; i++)
    delete_mark(get_nth_mark(0));

  /* and add the new ones */
  for (i=0; i<m_size; i++) {
    mark *nm;
    if (!(nm=new_mark(m_list[i]->name, m_list[i]->sample)))
      goto error_add;
    if (!m_list[i]->on)
      mark_off(nm);
  }

  /* and free local mem for marks */
  for (i=0; i<m_size; i++) {
    free(m_list[i]->name);
    free(m_list[i]);
  }
  if (m_maxsize) {
    free(m_list);
    m_list = 0;
    m_size = m_maxsize = 0;
  }

  modify = 0;
  return 0;

error:
  free_buffer();

  delete_channel_and_sub(mixer_channel);

  if (f)
    fclose(f);

  for (i=0; i<track_list_size; i++) {
    delete_channel_and_sub(track_list[i]->channel);
    free(track_list[i]);
  }
  if (track_list_maxsize) {
    free(track_list);
    track_list = 0;
    track_list_size = track_list_maxsize = 0;
  }

  for (i=0; i<track_names_size; i++)
    free(track_names[i]);
  if (track_names_maxsize) {
    free(track_names);
    track_names = 0;
    track_names_size = track_names_maxsize = 0;
  }

  for (i=0; i<m_size; i++) {
    if (m_list[i]->name)
      free(m_list[i]->name);
    free(m_list[i]);
  }
  if (m_maxsize) {
    free(m_list);
    m_list = 0;
    m_size = m_maxsize = 0;
  }

  modify = save_modify;
  return -1;

error_add:
  reset_error();
  add_error_char("You must quit the program.\n");
  add_error_char("The internal data is completely corrupted.\n");
  add_error_char("Do NOT save your current project, you will not lose any data,\n");
  add_error_char("because the error occured while loading a new project.\n");
  add_error_char("So, simply quit and restart.\n");
  add_error_char("And please, contact the authors and report this error message,\n");
  add_error_char("So that the problem can be solved.\n");
  display_error();
  return -1;

error0:
  reset_error();
  add_error_char("Load\n");
  add_error_char("The file ");
  add_error_char(current_file);
  add_error_char(" is not a valid Songs file\n");
  display_error();
  goto error;
}

static int print(FILE *f, char *c, ...)
{
  va_list ap;
  int ret;
  int serr;

  va_start(ap, c);
  ret = vfprintf(f, c, ap);
  serr = errno;
  va_end(ap);

  if (ret <= 0) {
    reset_error();
    add_error_char("Save\n");
    add_error_char(current_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    add_error_char("\n(the output file has been destroyed, solve the problem and save again)\n");
    display_error();
    return -1;
  }

  return 0;
}

/* we have to save the farther effect first */
int print_channels(FILE *f, channel *c)
{
  if (!c)
    return 0;

  if (print_channels(f, c->c.subchannel) == -1)
    return -1;

  switch (c->c.type) {
    case CHANNEL_TYPE_INPUT:
      if (print(f, "File ") == -1) return -1;
      if (((input *)c)->input_type == WAV_TYPE_MONO)
        if (print(f, "(wav mono) ") == -1) return -1;
      if (((input *)c)->input_type == WAV_TYPE_STEREO_LEFT)
        if (print(f, "(wav stereo left) ") == -1) return -1;
      if (((input *)c)->input_type == WAV_TYPE_STEREO_RIGHT)
        if (print(f, "(wav stereo right) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_16B_LITTLE_ENDIAN_MONO)
        if (print(f, "(raw 16b little endian mono) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT)
        if (print(f, "(raw 16b little endian stereo left) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT)
        if (print(f, "(raw 16b little endian stereo right) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_FLOAT_MONO)
        if (print(f, "(raw float mono) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_FLOAT_STEREO_LEFT)
        if (print(f, "(raw float stereo left) ") == -1) return -1;
      if (((input *)c)->input_type == RAW_TYPE_FLOAT_STEREO_RIGHT)
        if (print(f, "(raw float stereo right) ") == -1) return -1;
      if (print(f, "= %s", ((input *)c)->file) == -1) return -1;
      break;

    case CHANNEL_TYPE_POS:
      if (print(f, "Position = %d %d %d", ((pos *)c)->global_start, ((pos *)c)->subchannel_start, ((pos *)c)->subchannel_nb_samples) == -1) return -1;
      break;

    case CHANNEL_TYPE_PAN:
      if (print(f, "Panning = %lf", ((pan *)c)->pan) == -1) return -1;
      break;

    case CHANNEL_TYPE_VOL:
      if (print(f, "Gain = %lf", ((vol *)c)->vol_dB) == -1) return -1;
      break;

    case CHANNEL_TYPE_SC1:
      if (print(f, "Compressor = %lf %lf %lf %lf %lf %lf",
                ((sc1 *)c)->attack, ((sc1 *)c)->release, ((sc1 *)c)->threshold,
                ((sc1 *)c)->ratio, ((sc1 *)c)->knee, ((sc1 *)c)->makeup_gain)
            == -1) return -1;
      break;

    case CHANNEL_TYPE_MIXER:
      /* nothing to to */
      return 0;

    default:
      reset_error();
      add_error_char(__FILE__);
      add_error_char(":");
      add_error_int(__LINE__);
      add_error_char(":");
      add_error_char(__FUNCTION__);
      add_error_char(": there is a big problem,\n");
      add_error_char("save your data (if you can) and contact the authors.\n");
      display_error();
      return -1;
  }

  if (c->c.bypass) {
    if (print(f, " (bypass)") == -1)
      return -1;
  }

  if (print(f, "\n") == -1)
    return -1;

  return 0;
}

int save(char *name)
{
  FILE *f = fopen(name, "w");
  int i;
  int serr;

  current_file = name;

  if (!f) {
    serr = errno;
    reset_error();
    add_error_char("Save\n");
    add_error_char(current_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return -1;
  }

  if (print(f, "[Songs %s]\n", VERSION_SAVE_FILE) == -1) goto error;

  /* the markers */
  for (i=0; i<mark_list_size; i++) {
    if (print(f, "\n[Marker]") == -1) goto error;
    if (!mark_list[i]->on)
      if (print(f, " (hide)") == -1) goto error;
    if (print(f, "\n") == -1) goto error;
    if (print(f, "Name = %s\n", mark_list[i]->name) == -1) goto error;
    if (print(f, "Position = %d %d\n", mark_list[i]->sample, mark_list[i]->y_position) == -1) goto error;
  }

  /* the mixer settings */
  if (print(f, "\n[Mixer]\n") == -1) goto error;
  if (print_channels(f, main_mixer->mixer_track.channel) == -1) goto error;

  /* the tracks */
  for (i=0; i<main_mixer->nb_tracks; i++) {
    track *t = main_mixer->tracks[i];

    if (print(f, "\n[Track]") == -1) goto error;
    if (t->mute)
      if (print(f, " (mute)") == -1) goto error;
    if (print(f, "\n") == -1) goto error;

    if (print(f, "Name = %s\n", get_track_name(t)) == -1) goto error;

    if (print_channels(f, t->channel) == -1) goto error;
  }

  if (fclose(f) == EOF) {
    serr = errno;
    reset_error();
    add_error_char("Save\n");
    add_error_char(current_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    unlink(current_file);
    return -1;
  }

  modify = 0;
  return 0;

error:
  fclose(f);
  unlink(current_file);
  return -1;
}
