/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "mixer.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>

#include "error.h"
#include "vol.h"
#include "mix_gui.h"
#include "audio.h"

mixer *main_mixer;

static float mix_buf[4096*2];
static float mix_tmp[4096*2];

jmp_buf bus_error_handler;

static void bus_error(int unused)
{
  static int called = 0;

  signal(SIGBUS, bus_error);

  if (called == 0) {
    called = 1;
    reset_error();
    add_error_char("BUS ERROR\n");
    add_error_char("You should quit the program right now.\n");
    add_error_char("Try to save your data.\n");
    add_error_char("Contact the authors and tell them how you came there.\n");
    add_error_char("(This message won't show up again. It does not mean that there are no more bus errors. You really should quit now.)\n");
    display_error();
  }

  longjmp(bus_error_handler, 1);
}

/* return -1 if over */
static int mixer_16b_stereo_mix(mixer *m, short *buf, int nb_samples)
{
  static int called = 0;
  int i;
  int j;
  int current = m->next_sample;

  if (m->has_last_sample) {
    int nb_max = m->last_sample - current + 1;
    if (nb_max < nb_samples)
      nb_samples = nb_max;
  }

  if (nb_samples <= 0)
    return -1;

  if (setjmp(bus_error_handler)) {
    close_audio(); /* to be a little safe */
    return -1;
  }

  m->next_sample += nb_samples;

  if (nb_samples > 4096) {
    if (!called) {
      called = 1;
      reset_error();
      add_error_char(__FILE__);
      add_error_char(":");
      add_error_int(__LINE__);
      add_error_char(":");
      add_error_char(__FUNCTION__);
      add_error_char(": you should not read this, contact the authors.\n");
      display_error();
    }
    memset(buf, 0, sizeof(short) * 2 * nb_samples);
    return -1;
  }

  for (i=0; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (m->tracks[i]->channel->c.feed_buffer_float_stereo(m->tracks[i]->channel, mix_buf, nb_samples, current))
      break;
  }

#if 0
  /* no unmuted channel or no data */
  if (i == m->nb_tracks) {
    memset(buf, 0, sizeof(short) * 2 * nb_samples);

    /* preview requested ? */
    mix_preview_update(0, 0);

    return 0;
  }
#else
  if (i == m->nb_tracks)
    memset(mix_buf, 0, sizeof(float) * 2 * nb_samples);
#endif

  for (i++; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (!m->tracks[i]->channel->c.feed_buffer_float_stereo(m->tracks[i]->channel, mix_tmp, nb_samples, current))
      continue;

    for (j=0; j<nb_samples*2; j++)
      mix_buf[j] += mix_tmp[j];
  }

  /* let's now apply the mixer effects */
  main_mixer->mixer_track.channel->c.feed_buffer_float_stereo(main_mixer->mixer_track.channel, mix_buf, nb_samples, current);

  /* preview requested ? */
  mix_preview_update(mix_buf, nb_samples);

  /* and cut it for sending to the 16b audio card */
  for (i=0; i<nb_samples*2; i++) {
    register float z = mix_buf[i];
    if (z>32767) z = 32767;
    else if (z<-32768) z = -32768;
    buf[i] = z;
  }

  return 0;
}

/* returns -1 if over */
static int mixer_16b_mono_mix(mixer *m, short *buf, int nb_samples)
{
  static int called = 0;
  int i;
  int j;
  int current = m->next_sample;

  if (m->has_last_sample) {
    int nb_max = m->last_sample - current + 1;
    if (nb_max < nb_samples)
      nb_samples = nb_max;
  }

  if (nb_samples <= 0)
    return -1;

  if (setjmp(bus_error_handler)) {
    close_audio(); /* to be a little safe */
    return -1;
  }

  m->next_sample += nb_samples;

  if (nb_samples > 4096) {
    if (!called) {
      called = 1;
      reset_error();
      add_error_char(__FILE__);
      add_error_char(":");
      add_error_int(__LINE__);
      add_error_char(":");
      add_error_char(__FUNCTION__);
      add_error_char(": you should not read this, contact the authors.\n");
      display_error();
    }
    memset(buf, 0, sizeof(short) * nb_samples);
    return -1;
  }

  for (i=0; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (m->tracks[i]->channel->c.feed_buffer_float_mono(m->tracks[i]->channel, mix_buf, nb_samples, current))
      break;
  }

#if 0
  /* no unmuted channel or no data */
  if (i == m->nb_tracks) {
    memset(buf, 0, sizeof(short) * nb_samples);
    return 0;
  }
#else
  if (i == m->nb_tracks)
    memset(mix_buf, 0, sizeof(float) * nb_samples);
#endif

  for (i++; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (!m->tracks[i]->channel->c.feed_buffer_float_mono(m->tracks[i]->channel, mix_tmp, nb_samples, current))
      continue;

    for (j=0; j<nb_samples; j++)
      mix_buf[j] += mix_tmp[j];
  }

  /* let's now apply the mixer effects */
  main_mixer->mixer_track.channel->c.feed_buffer_float_mono(main_mixer->mixer_track.channel, mix_buf, nb_samples, current);

  /* and cut for the 16b audio card */
  for (i=0; i<nb_samples; i++) {
    register float z = mix_buf[i];
    if (z>32767) z = 32767;
    else if (z<-32768) z = -32768;
    buf[i] = z;
  }

  return 0;
}

/* return -1 if over */
static int mixer_float_stereo_mix(mixer *m, float *buf, int nb_samples)
{
  static int called = 0;
  int i;
  int j;
  int current = m->next_sample;

  if (m->has_last_sample) {
    int nb_max = m->last_sample - current + 1;
    if (nb_max < nb_samples)
      nb_samples = nb_max;
  }

  if (nb_samples <= 0)
    return -1;

  if (setjmp(bus_error_handler)) {
    close_audio(); /* to be a little safe */
    return -1;
  }

  m->next_sample += nb_samples;

  if (nb_samples > 4096) {
    if (!called) {
      called = 1;
      reset_error();
      add_error_char(__FILE__);
      add_error_char(":");
      add_error_int(__LINE__);
      add_error_char(":");
      add_error_char(__FUNCTION__);
      add_error_char(": you should not read this, contact the authors.\n");
      display_error();
    }
    memset(buf, 0, sizeof(float) * 2 * nb_samples);
    return -1;
  }

  for (i=0; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (m->tracks[i]->channel->c.feed_buffer_float_stereo(m->tracks[i]->channel, mix_buf, nb_samples, current))
      break;
  }

#if 0
  /* no unmuted channel or no data */
  if (i == m->nb_tracks) {
    memset(buf, 0, sizeof(float) * 2 * nb_samples);

    /* preview requested ? */
    mix_preview_update(0, 0);

    return 0;
  }
#else
  if (i == m->nb_tracks)
    memset(mix_buf, 0, sizeof(float) * 2 * nb_samples);
#endif

  for (i++; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (!m->tracks[i]->channel->c.feed_buffer_float_stereo(m->tracks[i]->channel, mix_tmp, nb_samples, current))
      continue;

    for (j=0; j<nb_samples*2; j++)
      mix_buf[j] += mix_tmp[j];
  }

  /* let's now apply the mixer effects */
  main_mixer->mixer_track.channel->c.feed_buffer_float_stereo(main_mixer->mixer_track.channel, mix_buf, nb_samples, current);

  /* preview requested ? */
  mix_preview_update(mix_buf, nb_samples);

  memcpy(buf, mix_buf, sizeof(float) * 2 * nb_samples);

  return 0;
}

/* return -1 if over */
static int mixer_float_mono_mix(mixer *m, float *buf, int nb_samples)
{
  static int called = 0;
  int i;
  int j;
  int current = m->next_sample;

  if (m->has_last_sample) {
    int nb_max = m->last_sample - current + 1;
    if (nb_max < nb_samples)
      nb_samples = nb_max;
  }

  if (nb_samples <= 0)
    return -1;

  if (setjmp(bus_error_handler)) {
    close_audio(); /* to be a little safe */
    return -1;
  }

  m->next_sample += nb_samples;

  if (nb_samples > 4096) {
    if (!called) {
      called = 1;
      reset_error();
      add_error_char(__FILE__);
      add_error_char(":");
      add_error_int(__LINE__);
      add_error_char(":");
      add_error_char(__FUNCTION__);
      add_error_char(": you should not read this, contact the authors.\n");
      display_error();
    }
    memset(buf, 0, sizeof(float) * nb_samples);
    return -1;
  }

  for (i=0; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (m->tracks[i]->channel->c.feed_buffer_float_mono(m->tracks[i]->channel, mix_buf, nb_samples, current))
      break;
  }

#if 0
  /* no unmuted channel or no data */
  if (i == m->nb_tracks) {
    memset(buf, 0, sizeof(float) * nb_samples);
    return 0;
  }
#else
  if (i == m->nb_tracks)
    memset(mix_buf, 0, sizeof(float) * nb_samples);
#endif

  for (i++; i < m->nb_tracks; i++) {
    if (m->tracks[i]->mute)
      continue;

    if (!m->tracks[i]->channel->c.feed_buffer_float_mono(m->tracks[i]->channel, mix_tmp, nb_samples, current))
      continue;

    for (j=0; j<nb_samples; j++)
      mix_buf[j] += mix_tmp[j];
  }

  /* let's now apply the mixer effects */
  main_mixer->mixer_track.channel->c.feed_buffer_float_mono(main_mixer->mixer_track.channel, mix_buf, nb_samples, current);

  memcpy(buf, mix_buf, sizeof(float) * nb_samples);

  return 0;
}

static void reset(void)
{
  /* do nothing */
}

static void reset_effects(mixer *m)
{
  int i;

  for (i=0; i<m->nb_tracks; i++)
    m->tracks[i]->channel->c.reset(m->tracks[i]->channel);

  m->mixer_track.channel->c.reset(m->mixer_track.channel);
}

/* return -1 on error */
int mixer_add_track(mixer *m, track *t)
{
  track **new;

  if (m->nb_tracks == m->max_nb_tracks) {
    new = realloc(m->tracks, (m->max_nb_tracks + 32) * sizeof(track *));
    if (!new) {
      oom();
      display_error();
      return -1;
    }
    m->tracks = new;
    m->max_nb_tracks += 32;
  }

  m->tracks[m->nb_tracks] = t;
  t->track_number = m->nb_tracks;

  m->nb_tracks++;

  return 0;
}

/* return -1 on error */
int mixer_sub_track(mixer *m, track *c)
{
  track **new;
  int i;

  i = c->track_number;

  if (m->tracks[i] != c) {
    reset_error();
    add_error_char("Someone did a bad hack inside Songs:\n");
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_char(__FUNCTION__);
    add_error_char("\nContact the authors\n");
    display_error();
    return -1;
  }

  for (; i < m->nb_tracks-1; i++) {
    m->tracks[i] = m->tracks[i+1];
    renumber_track(m->tracks[i], i);
  }

  m->nb_tracks--;

  /* let's try to free some memory */
  if (m->max_nb_tracks - m->nb_tracks > 64) {
    new = realloc(m->tracks, (m->max_nb_tracks - 32) * sizeof(track *));
    /* if it failed for a reason or another, we don't care, just return */
    if (!new)
      return 0;
    m->tracks = new;
    m->max_nb_tracks -= 32;
  }

  return 0;
}

/* this function is called by the effects set for the mixer, so it simply has to say yes,
 * so that the effects can be applied to the mixed audio stuff.
 */
static int do_nothing(void)
{
  return 1;
}

mixer *new_mixer(int type)
{
  mixer *m;

  if (type != MIXER_TYPE_16B_STEREO_INTERNAL_FLOAT) {
    reset_error();
    add_error_char("Unkown mixer type ");
    add_error_int(type);
    add_error_char(",\nyou should contact the authors.\n");
    display_error();
    return 0;
  }

  /* there may be bus errors if an audio file is resized and new size < old size, and we don't want Songs to crash */
  signal(SIGBUS, bus_error);

  m = malloc(sizeof(mixer));
  if (!m) {
    oom();
    display_error();
    return 0;
  }

  memset(m, 0, sizeof(mixer));

  m->mix = (int (*)(mixer *, void *, int))mixer_16b_stereo_mix;
  m->float_mix = (int (*)(mixer *, void *, int))mixer_float_stereo_mix;
  m->float_mono_mix = (int (*)(mixer *, void *, int))mixer_float_mono_mix;
  m->mix_mono = (int (*)(mixer *, void *, int))mixer_16b_mono_mix;
  m->reset_effects = reset_effects;

  m->_raw_mix.c.type = CHANNEL_TYPE_MIXER;

  m->_raw_mix.c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))do_nothing;
  m->_raw_mix.c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))do_nothing;
  m->_raw_mix.c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))do_nothing;
  m->_raw_mix.c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))do_nothing;
  m->_raw_mix.c.feed_buffer_short_n_channels = (int (*)(int, void *, short *, int, int))do_nothing;
  m->_raw_mix.c.feed_buffer_float_n_channels = (int (*)(int, void *, float *, int, int))do_nothing;
  m->_raw_mix.c.reset = (void (*)(void *))reset;

  m->mixer_track.channel = (channel *)new_vol(&m->_raw_mix);
  m->mixer_track.track_number = -1;

  if (!m->mixer_track.channel) {
    free(m);
    return 0;
  }

  return m;
}

void delete_mixer(mixer *m)
{
  free(m->tracks);
  free(m);
}

track *mixer_get_track(mixer *m, int t)
{
  return m->tracks[t];
}

int first_audio_sample(void)
{
  int i;
  int first = 0;
  int n;
  int all_muted = 1;

  if (main_mixer->nb_tracks == 0)
    return 0;

  if (!main_mixer->tracks[0]->mute) {
    all_muted = 0;
    first = main_mixer->tracks[0]->channel->c.first_point(main_mixer->tracks[0]->channel);
  }

  for (i=1; i<main_mixer->nb_tracks; i++) {
    if (main_mixer->tracks[i]->mute)
      continue;

    n = main_mixer->tracks[i]->channel->c.first_point(main_mixer->tracks[i]->channel);
    if (all_muted || n < first)
      first = n;

    all_muted = 0;
  }

  return first;
}

int last_audio_sample(void)
{
  int i;
  int last = 0;
  int n;
  int all_muted = 1;

  if (main_mixer->nb_tracks == 0)
    return 0;

  if (!main_mixer->tracks[0]->mute) {
    all_muted = 0;
    last = main_mixer->tracks[0]->channel->c.last_point(main_mixer->tracks[0]->channel);
  }

  for (i=1; i<main_mixer->nb_tracks; i++) {
    if (main_mixer->tracks[i]->mute)
      continue;

    n = main_mixer->tracks[i]->channel->c.last_point(main_mixer->tracks[i]->channel);
    if (all_muted || n > last)
      last = n;

    all_muted = 0;
  }

  return last;
}
