/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "raw.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "error.h"
#include "endian.h"
#include "memmap.h"

static void open_stream(input *c)
{
  if (c->stream)
    return;

  c->stream = fopen(c->file, "r");
}

static void close_stream(input *c)
{
  if (!c->stream)
    return;

  fclose((FILE *)c->stream);

  c->stream = 0;
}

static float buf[4096*2];

static void float_mono_float_stream(input *c, float *b, int nb_samples)
{
  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(b, nb_samples*4, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }
}

static void float_stereo_left_float_stream(input *c, float *b, int nb_samples)
{
  int i;
  float *bu = buf;

  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(buf, nb_samples*4*2, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }

  for (i=0; i<nb_samples; i++, b++, bu+=2)
    *b = *bu;
}

static void float_stereo_right_float_stream(input *c, float *b, int nb_samples)
{
  int i;
  float *bu = buf;

  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(buf, nb_samples*4*2, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }

  bu++;

  for (i=0; i<nb_samples; i++, b++, bu+=2)
    *b = *bu;
}

/* return 0 if error */
input *new_raw(char *file, int rate, int type)
{
  extern void wav_mono_float_stream(input *, float *, int);
  extern void wav_stereo_left_float_stream(input *, float *, int);
  extern void wav_stereo_right_float_stream(input *, float *, int);
  input tmp;
  input *ret;
  struct stat stat;
  int f;
  int serr;
  int bytes_per_sample;

  memset(&tmp, 0, sizeof(input));

  if (rate != 44100) {
    reset_error();
    add_error_char(file);
    add_error_char(":\nThe sample rate ");
    add_error_int(rate);
    add_error_char(" is not supported.\n");
    display_error();
    return 0;
  }

  if (type != RAW_TYPE_16B_LITTLE_ENDIAN_MONO         &&
      type != RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT  &&
      type != RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT &&
      type != RAW_TYPE_FLOAT_MONO                     &&
      type != RAW_TYPE_FLOAT_STEREO_LEFT              &&
      type != RAW_TYPE_FLOAT_STEREO_RIGHT) {
    reset_error();
    add_error_char(file);
    add_error_char(":\nThe type ");
    add_error_int(type);
    add_error_char(" is not supported.\n");
    display_error();
    return 0;
  }

  f = open(file, O_RDONLY);
  if (f == -1) {
    serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(":\n");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    return 0;
  }

  if (fstat(f, &stat) == -1) {
    serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(":\n");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    close(f);
    return 0;
  }

  bytes_per_sample = 0;
  switch(type) {
    case RAW_TYPE_16B_LITTLE_ENDIAN_MONO:
      bytes_per_sample = 2;
      break;
    case RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT:
    case RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT:
      bytes_per_sample = 4;
      break;
    case RAW_TYPE_FLOAT_MONO:
      bytes_per_sample = 4;
      break;
    case RAW_TYPE_FLOAT_STEREO_LEFT:
    case RAW_TYPE_FLOAT_STEREO_RIGHT:
      bytes_per_sample = 8;
      break;
  }

  if (stat.st_size & (bytes_per_sample-1) || !stat.st_size) {
    reset_error();
    add_error_char(file);
    if (stat.st_size) {
      add_error_char(": the file size is wrong, it should be multiple of ");
      add_error_int(bytes_per_sample);
      add_error_char("\n");
    } else
      add_error_char(": the file is empty\n");
    display_error();
    close(f);
    return 0;
  }

  tmp.size = stat.st_size / bytes_per_sample;
  tmp.data = memmap(&stat, f, file);

  if (tmp.data == (void *)-1) {
    serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(":\n");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    close(f);
    return 0;
  }

  close(f);

  tmp.file = strdup(file);
  tmp.file_size = stat.st_size;

  ret = malloc(sizeof(input));

  if (!tmp.file || !ret) {
    if (tmp.file) free(tmp.file);
    if (ret) free(ret);
    oom();
    display_error();
    memunmap(tmp.data, tmp.file_size);
    return 0;
  }

  *ret = tmp;

  ret->c.type = CHANNEL_TYPE_INPUT;
  ret->input_type = type;

  ret->open_stream = open_stream;
  ret->close_stream = close_stream;

  switch (type) {
    case RAW_TYPE_16B_LITTLE_ENDIAN_MONO:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))mono_short_little_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))mono_short_little_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))mono_short_little_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))mono_short_little_feed_buffer_float_stereo;
      ret->float_stream = wav_mono_float_stream;
      break;

    case RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_left_short_little_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_left_short_little_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_left_short_little_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_left_short_little_feed_buffer_float_stereo;
      ret->float_stream = wav_stereo_left_float_stream;
      break;

    case RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_right_short_little_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_right_short_little_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_right_short_little_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_right_short_little_feed_buffer_float_stereo;
      ret->float_stream = wav_stereo_right_float_stream;
      break;

    case RAW_TYPE_FLOAT_MONO:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))mono_float_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))mono_float_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))mono_float_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))mono_float_feed_buffer_float_stereo;
      ret->float_stream = float_mono_float_stream;
      break;

    case RAW_TYPE_FLOAT_STEREO_LEFT:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_left_float_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_left_float_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_left_float_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_left_float_feed_buffer_float_stereo;
      ret->float_stream = float_stereo_left_float_stream;
      break;

    case RAW_TYPE_FLOAT_STEREO_RIGHT:
      ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_right_float_feed_buffer_short_mono;
      ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_right_float_feed_buffer_float_mono;
      ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_right_float_feed_buffer_short_stereo;
      ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_right_float_feed_buffer_float_stereo;
      ret->float_stream = float_stereo_right_float_stream;
      break;
  }

  ret->c.first_point = (int (*)(void *))input_first_point;
  ret->c.last_point = (int (*)(void *))input_last_point;

  ret->c.reset = (void (*)(void *))input_reset;

  return ret;
}
