/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "record_gui.h"

#include <gtk/gtk.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include "soundcard.h"
#include "error.h"
#include "file_gui.h"
#include "main_gui.h"
#include "audio.h"
#include "mixer.h"
#include "input.h"
#include "version.h"
#include "pos.h"

int record_file_fd;

static char *output_file;
static int requested_record_mode;

static int record_start_sample;

static GtkWidget *win;
static GtkWidget *box_settings;
static GtkWidget *box_buttons;
static GtkWidget *filename_entry;
static GtkWidget *start_point_entry;
static GtkWidget *go_button;
static GtkWidget *stop_button;
static GtkWidget *play_toggle;
static GtkWidget *record_combo;
static GtkWidget *start_from_sample;

/* generate a file name that should not exist yet, not accurate */
static char out[64];
static void default_output_file(void)
{
  int i=1;
  struct stat stat_buf;

  /* We try to stat the file, if we get an error, then the file does
   * not exist. Of course, this is absolutely not accurate.
   * We hope the file name pattern is specific to Songs, so
   * we don't create too much problem with the rest of the world.
   */
  while(i<100000) {
    sprintf(out, "songs_record_track_%d", i); /* no buffer overflow */
    if (stat(out, &stat_buf) == -1)
      break;
    i++;
  }

  if (i == 100000) {
    reset_error();
    add_error_char("Record\n");
    add_error_char("Songs was unable to create a default output file name ");
    add_error_char("for the recording.\nYou will have to set a filename by hand.\n");
    display_error();
    output_file = 0;
    return;
  }

  output_file = out;
}

int in_record = 0;

static void set_file_name(char *name)
{
  struct stat stat_buf;
  static char name_buffer[FILENAME_MAX+1];
  int err = 0;

  if (strlen(name) > FILENAME_MAX) {
    reset_error();
    add_error_char("Record\n");
    add_error_char(name);
    add_error_char("\nThis filename is too long, Songs can't handle it (");
    add_error_int(FILENAME_MAX);
    add_error_char(" is the maximum).\n");
    display_error();
    err = 1;
  }

  /* if the file already exists, we don't overwrite it */
  if (!err && stat(name, &stat_buf) != -1) {
    reset_error();
    add_error_char("Record\n");
    add_error_char(name);
    add_error_char("\nThis file already exists. Songs will not overwrite it.\n");
    display_error();
    err = 1;
  }

  if (!err) {
    strcpy(name_buffer, name);
    output_file = name_buffer;
  } else
    default_output_file();

  gtk_entry_set_text(GTK_ENTRY(filename_entry), output_file);
  gtk_editable_set_position(GTK_EDITABLE(filename_entry), -1);
}

static void chose_file_event(void)
{
  if (in_record == 1) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char(__FUNCTION__);
    add_error_char(": you should not read this, contact the authors.\n");
    display_error();
    return;
  }

  gtk_widget_set_sensitive(win, FALSE);

  new_file_selector("Choose a file where to save recorded audio\n(the file name must be new):\n", set_file_name, win);

  gtk_widget_set_sensitive(win, TRUE);
}

static void set_record_mode_event(void)
{
  if (in_record == 1) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char(__FUNCTION__);
    add_error_char(": you should not read this, contact the authors.\n");
    display_error();
    return;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(play_toggle)))
    requested_record_mode = PLAY_AND_RECORD_SOUND;
  else
    requested_record_mode = RECORD_SOUND;
}

static void go_event(void)
{
  int ret;

  if (in_record == 1)
    return;

  /* let's calculate record_start_sample */
  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(start_from_sample)) == TRUE) {
    char *end;
    char *c = (char *)gtk_entry_get_text(GTK_ENTRY(start_point_entry));
    record_start_sample = strtol(c, &end, 10);
    if (*c == 0 || *end != 0) {
      reset_error();
      add_error_char("'");
      add_error_char(c);
      add_error_char("' is not a valid sample (it must be a number).\n");
      display_error();
      return;
    }
  } else {
    mark *m = get_mark((char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry)));
    if (!m) {
      record_start_sample = first_audio_sample();
    } else {
      record_start_sample = m->sample;
    }
  }

  if (output_file == 0) {
    reset_error();
    add_error_char("Record");
    add_error_char("No output file specified. Choose one.\n");
    display_error();
    return;
  }

  record_file_fd = open(output_file, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  if (record_file_fd == -1) {
    int serr = errno;
    reset_error();
    add_error_char("Record\n");
    add_error_char(output_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    add_error_char("\nHint: try another output file.\n");
    display_error();
    return;
  }

  gtk_widget_set_sensitive(box_settings, FALSE);
  gtk_widget_set_sensitive(box_buttons, FALSE);
  gtk_grab_add(win);


  ret = open_audio_play(requested_record_mode);

  if (ret == -1 || ret == -2) {
    close(record_file_fd);
    unlink(output_file);
    gtk_widget_set_sensitive(box_settings, TRUE);
    gtk_widget_set_sensitive(box_buttons, TRUE);
    gtk_grab_remove(win);
    return;
  }

  main_mixer->has_last_sample = 0;

  main_mixer->next_sample = record_start_sample;

  gtk_widget_grab_focus(stop_button);

  in_record = 1;
}

static void stop_event(void)
{
  if (in_record == 0)
    return;

  close_audio();

  in_record = 0;

  if (close(record_file_fd) == -1) {
    int serr = errno;
    reset_error();
    add_error_char("Record\n");
    add_error_char(output_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    add_error_char("(the file is not destroyed, so you can try to do something with it)\n");
    display_error();
  }

  gtk_widget_set_sensitive(box_settings, TRUE);
  gtk_widget_set_sensitive(box_buttons, TRUE);
  gtk_grab_remove(win);
}

static int record_gui_on = 0;

static void done_event(void)
{
  track *t;

  if (in_record == 1) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char(__FUNCTION__);
    add_error_char(": you should not read this, contact the authors\n");
    display_error();
    return;
  }

  t = new_track(output_file, RAW_TYPE_16B_LITTLE_ENDIAN_MONO);
  if (t)
    set_global_start((pos *)t->channel, record_start_sample);

  record_gui_on = 0;
}

static void cancel_event(void)
{
  if (in_record == 1) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char(__FUNCTION__);
    add_error_char(": you should not read this, contact the authors\n");
    display_error();
    return;
  }

  if (output_file)
    unlink(output_file);

  record_gui_on = 0;
}

static int delete_event(void)
{
  if (in_record)
    stop_event();

  cancel_event();

  return FALSE;

  //main_gui_activate_menu();
}

static int adjust_in = 0;
short adjust_audio_buf[2048];
static unsigned char adjust_min[256];
static unsigned char adjust_max[256];
static unsigned char adjust_next_point;
static GtkWidget *adjust_input_draw;

static int draw_expose_event(GtkWidget *draw, GdkEventExpose *ev)
{
  int len = ev->area.width;
  int i;
  for (i=ev->area.x; len; len--, i++)
    gdk_draw_line(ev->window, draw->style->black_gc, i, adjust_min[i], i, adjust_max[i]);

  return TRUE;
}

void adjust_input_update(void)
{
  int min, max;
  int i;

  if (adjust_next_point == 0) {
    for (i=0; i<256; i++)
      adjust_min[i] = adjust_max[i] = 127;
    gtk_widget_queue_draw(adjust_input_draw);
  }

  min = max = adjust_audio_buf[0];

  for (i=1; i<2048; i++) {
    if (adjust_audio_buf[i] < min) min = adjust_audio_buf[i];
    if (adjust_audio_buf[i] > max) max = adjust_audio_buf[i];
  }

  /*  -32768    x     32767
   *    255     y       0
   * y / 255 = (x-32767) / (-32768-32767)
   */
  adjust_min[adjust_next_point] = ((float)min-32767.) * 255. / (-32768.-32767.);
  adjust_max[adjust_next_point] = ((float)max-32767.) * 255. / (-32768.-32767.);

  gtk_widget_queue_draw_area(adjust_input_draw, adjust_next_point, 0, 1, 256);

  adjust_next_point++;
}

static void adjust_input_destroy(void)
{
  close_audio();
  adjust_in = 0;
}

static void start_adjust_input(void)
{
  open_audio_play(RECORD_ADJUST);
}

static GtkWidget *awin;

static void adjust_input_event(void)
{
  GtkWidget *draw;
  GtkWidget *vbox, *hbox;
  GtkWidget *go, *stop, *cancel;
  int i;

  if (adjust_in)
    return;

  /* data */
  adjust_next_point = 0;
  for (i=0; i<256; i++)
    adjust_min[i] = adjust_max[i] = 127;

  /* gui */
  awin = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(awin), 5);
  /* error used as buffer */
  reset_error();
  add_error_char("Songs ");
  add_error_char(VERSION);
  add_error_char(" - Adjust Input Level");
  gtk_window_set_title(GTK_WINDOW(awin), get_error());

  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(awin), vbox);

  adjust_input_draw = draw = gtk_drawing_area_new();
  gtk_widget_modify_bg(draw, GTK_STATE_NORMAL, &draw->style->white);
  gtk_widget_set_size_request(draw, 256, 256);
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), draw, TRUE, FALSE, 0);

  go = gtk_button_new_with_label("Go");
  stop = gtk_button_new_with_label("Stop");
  cancel = gtk_button_new_with_label("Cancel");
  hbox = gtk_hbox_new(TRUE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), go, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), stop, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), cancel, TRUE, TRUE, 0);

  /* callbacks */
  g_signal_connect(awin, "destroy", G_CALLBACK(adjust_input_destroy), 0);
  g_signal_connect_swapped(cancel, "clicked", G_CALLBACK(gtk_widget_destroy), awin);
  g_signal_connect(go, "clicked", G_CALLBACK(start_adjust_input), 0);
  g_signal_connect(stop, "clicked", G_CALLBACK(close_audio), 0);
  g_signal_connect(draw, "expose-event", G_CALLBACK(draw_expose_event), 0);

  gtk_widget_show_all(awin);

  adjust_in = 1;
}

static void record_off(void)
{
  if (adjust_in)
    gtk_widget_destroy(awin);
}

void record_gui(void)
{
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *filename_button;
  GtkWidget *ok;
  GtkWidget *cancel;
  GtkWidget *start_from_mark;
  GtkWidget *adjust_input;

  if (record_gui_on)
    return;

  requested_record_mode = RECORD_SOUND;

  default_output_file();

  /* the interface */
  win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(win), 5);
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(win), vbox);

  /* error used as buffer */
  reset_error();
  add_error_char("Songs ");
  add_error_char(VERSION);
  add_error_char(" - record");
  gtk_window_set_title(GTK_WINDOW(win), get_error());

  /* the settings stuff */
  box_settings = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), box_settings, FALSE, FALSE, 0);
  /* filename */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(box_settings), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Output file"), FALSE, FALSE, 0);
  filename_entry = gtk_entry_new();
  gtk_entry_set_editable(GTK_ENTRY(filename_entry), FALSE);
  gtk_entry_set_text(GTK_ENTRY(filename_entry), output_file);
  gtk_editable_set_position(GTK_EDITABLE(filename_entry), -1);
  gtk_box_pack_start(GTK_BOX(hbox), filename_entry, TRUE, TRUE, 0);
  filename_button = gtk_button_new_with_label("Choose");
  gtk_box_pack_start(GTK_BOX(hbox), filename_button, FALSE, FALSE, 0);
  /* start point */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(box_settings), hbox, FALSE, FALSE, 0);
  start_from_sample = gtk_radio_button_new_with_label(0, "Start from sample");
  gtk_box_pack_start(GTK_BOX(hbox), start_from_sample, FALSE, FALSE, 0);
  start_point_entry = gtk_entry_new();
  gtk_entry_set_width_chars(GTK_ENTRY(start_point_entry), 10);
  gtk_entry_set_text(GTK_ENTRY(start_point_entry), "0");
  gtk_box_pack_start(GTK_BOX(hbox), start_point_entry, TRUE, TRUE, 0);
  /* combo */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(box_settings), hbox, FALSE, FALSE, 0);
  start_from_mark = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(start_from_sample), "Start from mark");
  gtk_box_pack_start(GTK_BOX(hbox), start_from_mark, FALSE, FALSE, 0);
  record_combo = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(hbox), record_combo, TRUE, TRUE, 0);
  /* record mode */
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(box_settings), hbox, FALSE, FALSE, 0);
  play_toggle = gtk_check_button_new_with_mnemonic("_Play while record");
  gtk_box_pack_start(GTK_BOX(hbox), play_toggle, FALSE, FALSE, 0);
  /* adjust input line */
  adjust_input = gtk_button_new_with_label("Adjust input level");
  gtk_box_pack_end(GTK_BOX(hbox), adjust_input, FALSE, FALSE, 0);

  /* the Go / Stop buttons */
  go_button = gtk_button_new_with_mnemonic("_Go");
  stop_button = gtk_button_new_with_mnemonic("_Stop");
  hbox = gtk_hbox_new(TRUE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), go_button, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), stop_button, FALSE, FALSE, 0);
  GTK_WIDGET_SET_FLAGS(stop_button, GTK_CAN_FOCUS);

  /* the Ok / Cancel buttons */
  box_buttons = gtk_hbox_new(FALSE, 10);
  gtk_box_pack_start(GTK_BOX(vbox), box_buttons, FALSE, FALSE, 0);
  ok = gtk_button_new_with_mnemonic("_OK");
  cancel = gtk_button_new_with_mnemonic("_Cancel");
  gtk_box_pack_end(GTK_BOX(box_buttons), ok, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(box_buttons), cancel, FALSE, FALSE, 0);

  /* callback stuff */
  g_signal_connect(filename_button, "clicked", G_CALLBACK(chose_file_event), 0);
  g_signal_connect(play_toggle, "toggled", G_CALLBACK(set_record_mode_event), 0);
  g_signal_connect(go_button, "clicked", G_CALLBACK(go_event), 0);
  g_signal_connect(stop_button, "clicked", G_CALLBACK(stop_event), 0);
  g_signal_connect(ok, "clicked", G_CALLBACK(done_event), 0);
  g_signal_connect(cancel, "clicked", G_CALLBACK(cancel_event), 0);

  g_signal_connect(adjust_input, "clicked", G_CALLBACK(adjust_input_event), 0);

  g_signal_connect_swapped(ok, "clicked", G_CALLBACK(gtk_widget_destroy), win);
  g_signal_connect_swapped(cancel, "clicked", G_CALLBACK(gtk_widget_destroy), win);

  g_signal_connect(win, "delete-event", G_CALLBACK(delete_event), 0);
  g_signal_connect(win, "delete-event", G_CALLBACK(gtk_widget_destroy), 0);

  g_signal_connect(win, "destroy", G_CALLBACK(record_off), 0);

  gtk_widget_show_all(win);

  //main_gui_disactivate_menu();

  record_gui_on = 1;

  /* this must come after record_gui_on = 1 */
  record_gui_update_start_record_list();
}

void error_record(void)
{
  if (!in_record)
    return;

  stop_event();
}

static GList *record_list = 0;

void record_gui_update_start_record_list(void)
{
  GList *new;
  int i;

  if (!record_gui_on)
    return;

  if (record_list) {
    g_list_free(record_list);
    record_list = 0;
  }

  record_list = g_list_append(record_list, "First audio sample");
  if (!record_list) {
    oom();
    display_error();
    return;
  } 

  for (i=0; i<mark_list_size; i++) {
    new = g_list_append(record_list, mark_list[i]->name);
    if (!new) {
      oom();
      display_error();
      g_list_free(record_list);
      record_list = 0;
      return;
    }
    record_list = new;
  }

  if (!record_list) {
    record_list = g_list_append(0, "");
    if (!record_list) {
      oom();
      display_error();
      return;
    }
  }

  /* error used as buffer */
  reset_error();
  add_error_char((char *)gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry)));

  gtk_combo_set_popdown_strings(GTK_COMBO(record_combo), record_list);

  if (!get_mark(get_error())) {
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry), "");
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry), record_list->data);
  } else {
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry), "");
    gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(record_combo)->entry), get_error());
  }
}
