/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "soundcard.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "error.h"
#include "setup_gui.h"

int fragment_size[11]={         /* 7, ..., 17 makes 11 values */
  /* fragment size is in [7..17] says mikmod 3.0.3, so... */
  128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072
};

/* the audio device file handle */
static int soundd;

static int inited=0;

void sound_stats(void)
{
#if 0
  struct audio_buf_info in, out;
  int cap;
  if (ioctl(soundd, SNDCTL_DSP_GETOSPACE, &out) == -1 ||
      ioctl(soundd, SNDCTL_DSP_GETISPACE, &in)  == -1) {
    perror("ioctl");
    return;
  }
  fprintf(stdout, "%d %d in: %d %d %d %d out : %d %d %d %d\n",
    in.fragsize, out.fragsize,
    in.fragments, in.fragstotal, in.fragsize, in.bytes,
    out.fragments, out.fragstotal, out.fragsize, out.bytes);

  if (ioctl(soundd, SNDCTL_DSP_GETCAPS, &cap) == -1) {
    perror("ioctl");
    return;
  }
  fprintf(stdout, "%d\n", !!(cap & DSP_CAP_BATCH));
#endif

  struct count_info in, out;
  if (ioctl(soundd, SNDCTL_DSP_GETOPTR, &out) == -1 ||
      ioctl(soundd, SNDCTL_DSP_GETIPTR, &in)  == -1) {
    perror("ioctl");
    return;
  }

  fprintf(stdout, "in: bytes = %d blocks = %d ptr = %d\nout: bytes = %d blocks = %d ptr = %d\n",
     in.bytes, in.blocks, in.ptr,
     out.bytes, out.blocks, out.ptr);

  fprintf(stdout, "diff = %d\n", in.bytes - out.bytes * 2);
}

int get_sound_fd(void)
{
  if (!inited)
    return -1;

  return soundd;
}

/* return 0 if full (or error) 1 if not */
int not_full_sound(void)
{
  audio_buf_info info;

  if (ioctl(soundd, SNDCTL_DSP_GETOSPACE, &info)==-1)
    return 0;

  return info.fragments!=0;
}

/* this function sets the card output resolution, ie 8/16 bits, mono(0)/stereo(1), sampling rate (8000, 44100, ...) */
/* return 0 on error, 1 on success */
int set_sound_type(int bit_resolution, int mono_stereo, int sampling_rate, int fragsize)
{
  int numfrags=9;
  int s0, s1, s2;
  int fragmentsize;

  s0 = fragmentsize = ((numfrags<<16) | fragsize);

  if(ioctl(soundd, SNDCTL_DSP_SETFRAGMENT, &fragmentsize)<0) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("ioctl on ");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return 0;
  }

  if (s0 != fragmentsize)
    goto not_supported;

  s0 = bit_resolution;
  s1 = mono_stereo;
  s2 = sampling_rate;

  if (ioctl(soundd, SNDCTL_DSP_SAMPLESIZE, &bit_resolution)==-1 ||
      ioctl(soundd, SNDCTL_DSP_STEREO, &mono_stereo)==-1 ||
      ioctl(soundd, SNDCTL_DSP_SPEED, &sampling_rate)==-1) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("ioctl on ");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return 0;
  }

  if (s0 != bit_resolution || s1 != mono_stereo || s2 != sampling_rate)
    goto not_supported;

  ioctl(soundd, SNDCTL_DSP_RESET, 0);

  return 1;

not_supported:
  reset_error();
  add_error_char("Audio\n");
  add_error_char("The requested audio mode is not supported by your soundcard.\n");
  display_error();
  return 0;
}

/* return 0 if ok, -1 otherwise */
int unblock_sound(int fd)
{
  int fl;

  if ((fl=fcntl(fd, F_GETFL))==-1) {
    perror("card.unblock_sound.fcntl1");
    return -1;
  }
  fl|=O_NONBLOCK;
  if ((fl=fcntl(fd, F_SETFL, fl))==-1) {
    perror("card.unblock_soud.fcntl2");
    return -1;
  }
  return 0;
}

/* return 0 on error 1 otherwise*/
int init_sound(int sample_rate, int precision, int mono_stereo, int fragsize, int type)
{
  int mode = 0;

  if (inited==1) {
    reset_error();
    add_error_char("Audio\n");
    add_error_char("Audio already in use inside Songs\n");
    display_error();
    return 0;
  }

  if (type == PLAY_SOUND)
    mode = O_WRONLY;
  else if (type == RECORD_SOUND)
    mode = O_RDONLY;
  else if (type == PLAY_AND_RECORD_SOUND)
    mode = O_RDWR;
  else {
    reset_error();
    add_error_char("Audio\n");
    add_error_char("Bad mode for the audio stuff: ");
    add_error_int(type);
    add_error_char("\n");
    display_error();
    return 0;
  }

  soundd=open(audio_file, mode, 0);
  if (soundd == -1) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    if (type == PLAY_AND_RECORD_SOUND) {
      add_error_char("(Your soundcard probably does not support full duplex\n");
      add_error_char("or there is no driver for it in your operating system)\n");
    }
    display_error();
    return 0;
  }

  if (!set_sound_type(precision, mono_stereo, sample_rate, fragsize) /*|| unblock_sound(soundd)==-1*/) {
    close(soundd);
    return 0;
  }

  inited=1;

  return 1;
}

void close_sound(void)
{
  if (!inited)
    return;

  ioctl(soundd, SNDCTL_DSP_RESET, 0);
  close(soundd);
  inited=0;
}

/* return 0 if problem, 1 if ok */
int play_sound(void *buf, int buf_len)
{
  int serr;

  if (!inited)
    return 0;
  if (write(soundd, buf, buf_len)!=buf_len) {
    serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("write:");
    add_error_char(strerror(serr));
    display_error();
    close_sound();
    return 0;
  }
  return 1;
}

/* return 0 if problem, 1 if ok */
int record_sound(void *buf, int buf_len)
{
  int serr;

  if (!inited)
    return 0;
  if (read(soundd, buf, buf_len)!=buf_len) {
    serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("read:");
    add_error_char(strerror(serr));
    display_error();
    close_sound();
    return 0;
  }
  return 1;
}
