/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "track.h"

#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include "channel.h"
#include "error.h"
#include "input.h"
#include "wav.h"
#include "pos.h"
#include "pan.h"
#include "vol.h"
#include "mixer.h"
#include "track_gui.h"
#include "main_gui.h"
#include "effects_gui.h"
#include "file_gui.h"

static void auto_load(char *file)
{
  struct stat buf;
  int type;
  int type2 = 0 /* for no gcc warnings */;
  track *n1;
  track *n2;

  /* does the file exist ? */
  if (stat(file, &buf) == -1) {
    int serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return;
  }

  type = is_wav(file);

  if (type == -2) {
    reset_error();
    add_error_char(file);
    add_error_char(": unsupported wav file type (only 44100Hz 16b mono/stereo are handled)\n");
    display_error();
    return;
  }

  if (type == WAV_TYPE_STEREO_LEFT) {
    type2 = WAV_TYPE_STEREO_RIGHT;
    goto stereo_load;
  }

  if (type == WAV_TYPE_MONO) {
    new_track(file, WAV_TYPE_MONO);
    return;
  }

  /* so, not a wav */
  type = ask_user_file_type(file);
  if (type == -1)
    return;

  if (type == RAW_TYPE_16B_LITTLE_ENDIAN_MONO ||
      type == RAW_TYPE_FLOAT_MONO) {
    new_track(file, type);
    return;
  }

  if (type == RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_LEFT)
    type2 = RAW_TYPE_16B_LITTLE_ENDIAN_STEREO_RIGHT;

  if (type == RAW_TYPE_FLOAT_STEREO_LEFT)
    type2 = RAW_TYPE_FLOAT_STEREO_RIGHT;

stereo_load:
  n1 = new_track(file, type);
  if (!n1)
    return;

  n2 = new_track(file, type2);
  if (!n2) {
    delete_track(n1);
    return;
  }

  set_pan((pan *)n1->channel, -1);
  set_pan((pan *)n2->channel, 1);
}

track *new_track(char *file, int type)
{
  track *new;
  channel *in;
  pos *pos;
  vol *vol;
  pan *pan;

  /* get the file */
  if (type == INPUT_TYPE_AUTOMATIC) {
    auto_load(file);
    return 0;
  }

  in = (channel *)new_input(file, type);

  if (!in)
    return 0;

  pos = new_pos(in);
  if (!pos) {
    delete_channel(in);
    return 0;
  }

  pan = new_pan((channel *)pos);
  if (!pan) {
    delete_channel(in);
    delete_pos(pos);
    return 0;
  }

  vol = new_vol((channel *)pan);
  if (!vol) {
    delete_channel(in);
    delete_pos(pos);
    delete_pan(pan);
    return 0;
  }

  new = malloc(sizeof(track));

  if (!new) {
    reset_error();
    oom();
    display_error();
    delete_channel(in);
    delete_pos(pos);
    delete_pan(pan);
    delete_vol(vol);
    return 0;
  }

  memset(new, 0, sizeof(track));

  new->channel = (channel *)vol;
  new->input = (input *)in;

  if (mixer_add_track(main_mixer, new) == -1) {
    delete_channel(in);
    delete_pos(pos);
    delete_pan(pan);
    delete_vol(vol);
    free(new);
    return 0;
  }

  track_gui(new);

  main_gui_add_track(new);
  effects_gui_new_track(new);

  modify = 1;

  /* set a custom name */
  {
    char *f = file + strlen(file) - 1;
    while (f != file && *f != '/')
      f--;
    if (*f == '/') f++;

    /* error used as buffer */
    reset_error();
    add_error_char(f);
    if (type == WAV_TYPE_STEREO_LEFT)
      add_error_char(" (left)");
    if (type == WAV_TYPE_STEREO_RIGHT)
      add_error_char(" (right)");

    set_track_name(new, get_error());
  }

  return new;
}

void delete_track(track *t)
{
  channel *c = t->channel;

  mixer_sub_track(main_mixer, t);

  track_ungui(t);

  main_gui_sub_track(t);
  effects_gui_delete_track(t);

  delete_channel_and_sub(c);

  free(t);

  modify = 1;
}

void renumber_track(track *t, int n)
{
  channel *c;

  if (t->track_number == n)
    return;

  t->track_number = n;

  /* error used as a buffer */
  reset_error();
  add_error_char("# ");
  add_error_int(n + 1);

  set_label_text(t->label, get_error());

  c = t->channel;

  while(c) {
    if (c->c.gui_on && c->c.label)
      set_label_text(c->c.label, get_error());

    c = c->c.subchannel;
  }
}
