/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "view_gui.h"

#include <gtk/gtk.h>
#include <stdlib.h>

/* it seems to me that scanning the file with fopen/fread/fclose is better than directly using the mmaped data.
 * on a huge file (700MB) (but over nfs, so maybe not a very common case), it takes a little less time and much
 * much less memory (the kernel won't swap other processes to make room for the mmaped data)
 */
#define GENERATE_LOW_MEM

#ifndef GENERATE_LOW_MEM
#include <setjmp.h>
#endif

#include "track.h"
#include "channel.h"
#include "input.h"
#include "pos.h"
#include "error.h"
#include "mixer.h"
#include "main_gui.h"

static int view_used = 0;

static int user_interrupt;

static int nb_cols;
static int col_cur;
static int next_sample;
static int size;
static track *cur_track;
static input *input_channel;

static GtkWidget *win;
static GtkWidget *bar;

static float buf[4096];
static void do_calculation(void)
{
  int s;
  float min, max;
  int i;
  int nb_turns=10;
#ifndef GENERATE_LOW_MEM
  extern jmp_buf bus_error_handler;
#endif

#ifdef GENERATE_LOW_MEM
  input_channel->open_stream(input_channel);
#else
  /* we can get bus errors, so we handle them (not 100% sure code) */
  if (setjmp(bus_error_handler)) {
    free(cur_track->input->view_min);
    free(cur_track->input->view_max);
    cur_track->input->view_min = cur_track->input->view_max = 0;
    gtk_widget_destroy(win);
    return;
  }
#endif

one_more:
  s = size;

  if (s > 4096)
    s = 4096;

#ifdef GENERATE_LOW_MEM
  input_channel->float_stream(input_channel, buf, s); //, next_sample);
#else
  input_channel->c.feed_buffer_float_mono(input_channel, buf, s, next_sample);
#endif

  min = max = buf[0];

  for (i=1; i<s; i++) {
    if (buf[i] < min) min = buf[i];
    if (buf[i] > max) max = buf[i];
  }

  if (min < -32768) min = -32768;
  if (min > 32767) min = 32767;
  if (max < -32768) max = -32768;
  if (max > 32767) max = 32767;

  /* -32768        x    32767
   *   63          y      0
   * (y-63) / -63 = (x+32768) / (32767+32768)
   * y = 63 - (x+32768) * 63 / (32767 + 32768)
   */

  min = 63. - (min + 32768.) * 63. / (32767.+32768.);
  max = 63. - (max + 32768.) * 63. / (32767.+32768.);

  cur_track->input->view_min[col_cur] = min;
  cur_track->input->view_max[col_cur] = max;

  col_cur++;

  next_sample += 4096;

  size -= s;
  if (!size) {
    gtk_widget_destroy(win);
#ifdef GENERATE_LOW_MEM
    input_channel->close_stream(input_channel);
#endif
    return;
  }

  nb_turns--;
  if (nb_turns)
    goto one_more;

  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(bar), (float)col_cur / nb_cols);

  while (gtk_events_pending())
    gtk_main_iteration();

  if (user_interrupt) {
#ifdef GENERATE_LOW_MEM
    input_channel->close_stream(input_channel);
#endif
    free(cur_track->input->view_min);
    free(cur_track->input->view_max);
    cur_track->input->view_min = cur_track->input->view_max = 0;
    return;
  }

  nb_turns = 10;
  goto one_more;
}

static void destroy_event(void)
{
  user_interrupt = 1;
}

void generate_view(track *t)
{
  extern GtkWidget *main_win;
  GtkWidget *box;
  GtkWidget *hbox;
  GtkWidget *cancel;

  if (view_used)
    return;

  if (t->input->view_min)
    return;

  /* data */
  cur_track = t;

  input_channel = t->input;

  size = nb_cols = input_channel->c.last_point(input_channel);
  nb_cols += 4095;
  nb_cols /= 4096;
  col_cur = 0;

  input_channel->view_min = malloc(nb_cols);
  if (!input_channel->view_min) {
    reset_error();
    add_error_char("Not enough memory to generate view.\n");
    display_error();
    return;
  }

  input_channel->view_max = malloc(nb_cols);
  if (!input_channel->view_max) {
    free(input_channel->view_min);
    input_channel->view_min = 0;
    reset_error();
    add_error_char("Not enough memory to generate view.\n");
    display_error();
    return;
  }

  next_sample = 0;

  view_used = 1;

  user_interrupt = 0;

  /* interface */
  win = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  box = gtk_vbox_new(FALSE, 5);
  gtk_container_add(GTK_CONTAINER(win), box);
  gtk_container_set_border_width(GTK_CONTAINER(box), 5);

  gtk_box_pack_start(GTK_BOX(box), gtk_label_new("Generating view for the file:"), FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(box), gtk_label_new(((input *)input_channel)->file), FALSE, FALSE, 0);

  bar = gtk_progress_bar_new();
  gtk_box_pack_start(GTK_BOX(box), bar, FALSE, FALSE, 0);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(bar), 0);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
  cancel = gtk_button_new_with_mnemonic("_Cancel");
  gtk_box_pack_end(GTK_BOX(hbox), cancel, FALSE, FALSE, 0);

  /* signals */
  g_signal_connect(win, "destroy", G_CALLBACK(destroy_event), 0);
  g_signal_connect_swapped(cancel, "clicked", G_CALLBACK(gtk_widget_destroy), win);

  gtk_window_set_modal(GTK_WINDOW(win), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(win), GTK_WINDOW(main_win));
  gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);

  gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER_ON_PARENT);

  /* done */
  gtk_widget_show_all(win);

  do_calculation();

  main_gui_redraw_track_view();

  view_used = 0;
}

void generate_view_all(void)
{
  int i;

  for (i=0; i<main_mixer->nb_tracks; i++)
    generate_view(main_mixer->tracks[i]);

  main_gui_redraw_track_view();
}

void generate_view_none(void)
{
  int i;

  for (i=0; i<main_mixer->nb_tracks; i++) {
    track *t = main_mixer->tracks[i];
    if (!t->input->view_min) continue;
    free(t->input->view_min);
    free(t->input->view_max);
    t->input->view_min = t->input->view_max = 0;
  }

  main_gui_redraw_track_view();
}

gboolean expose_area_event(track *t, GdkEventExpose *ev)
{
  GdkGC *on;
  GdkGC *off;
  int i;
  int len = ev->area.width;
  int p;
  pos *ps;
  int start, end;

  ps = (pos *)t->channel;
  while (ps->c.type != CHANNEL_TYPE_POS)
    ps = (pos *)ps->c.subchannel;

  start = ps->subchannel_start;
  end = start + ps->subchannel_nb_samples-1;

  on = ((GtkWidget *)t->drawing_area)->style->black_gc;
  off = ((GtkWidget *)t->drawing_area)->style->mid_gc[0];

  if (!t->input->view_min) {
    for (i = ev->area.x, p = i * 4096; len; len--, i++, p+=4096) {
      register int black = p>=start && p<=end;
      black |= (p+4095)>=start && (p+4095)<=end;
      black |= p<start && (p+4095)>end;
      gdk_draw_line(ev->window, black ? on : off, i, 16, i, 48);
    }
    return TRUE;
  }

  for (i = ev->area.x, p = i * 4096; len; len--, i++, p+=4096) {
    register int black = p>=start && p<=end;
    black |= (p+4095)>=start && (p+4095)<=end;
    black |= p<start && (p+4095)>end;
    gdk_draw_line(ev->window, black ? on : off, i, t->input->view_min[i], i, t->input->view_max[i]);
  }

  return TRUE;
}

void view_set_callbacks(track *t)
{
  g_signal_connect_swapped(t->drawing_area, "expose-event", G_CALLBACK(expose_area_event), t);
}
