/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "wav.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "error.h"
#include "memmap.h"

static void open_stream(input *c)
{
  char b[44];
  FILE *f;

  if (c->stream)
    return;

  f = fopen(c->file, "r");

  if (!f)
    return;

  if (fread(b, 44, 1, f) != 1) {
    fclose(f);
    return;
  }

  c->stream = f;
}

static void close_stream(input *c)
{
  if (!c->stream)
    return;

  fclose((FILE *)c->stream);

  c->stream = 0;
}

static short buf[4096*2];

void wav_mono_float_stream(input *c, float *b, int nb_samples)
{
  int i;
  short *bu = buf;

  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(buf, nb_samples*2, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }

  for (i=0; i<nb_samples; i++, b++, bu++)
    *b = *bu;
}

void wav_stereo_left_float_stream(input *c, float *b, int nb_samples)
{
  int i;
  short *bu = buf;

  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(buf, nb_samples*2*2, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }

  for (i=0; i<nb_samples; i++, b++, bu+=2)
    *b = *bu;
}

void wav_stereo_right_float_stream(input *c, float *b, int nb_samples)
{
  int i;
  short *bu = buf;

  if (!c->stream || nb_samples > 4096) {
    memset(b, 0, nb_samples*4);
    return;
  }

  if (fread(buf, nb_samples*2*2, 1, (FILE *)c->stream) != 1) {
    fclose(c->stream);
    c->stream = 0;
    memset(b, 0, nb_samples*4);
    return;
  }

  bu++;

  for (i=0; i<nb_samples; i++, b++, bu+=2)
    *b = *bu;
}

static unsigned short mem_to_short(unsigned char *b)
{
  unsigned short t;

  t = *b++;
  t += ((*b++) << 8);

  return t;
}

static unsigned long mem_to_long(unsigned char *b)
{
  unsigned long t;

  t = *b++;
  t += ((*b++) << 8);
  t += ((*b++) << 16);
  t += ((*b) << 24);

  return t;
}

static void short_to_mem(char *b, unsigned short v)
{
  *b++ = v & 255;
  *b = (v >> 8) & 255;
}

static void long_to_mem(char *b, unsigned long v)
{
  *b++ = v & 255;
  *b++ = (v >> 8) & 255;
  *b++ = (v >> 16) & 255;
  *b = (v >> 24) & 255;
}

/* returns WAV_TYPE_MONO if mono file
 * WAV_TYPE_STEREO_LEFT if stereo.
 * the file must be 16b, 44100 Hz
 * return -1 if error
 * -2 if unsupported file type
 */
static int analyze_wav(char *data, struct stat *stat)
{
  int nb_channels;

  /* http://ccrma-www.stanford.edu/CCRMA/Courses/422/projects/WaveFormat/
   * explains the format.
   * (the url was ok Mon Dec 29 18:17:02 CET 2003)
   * here is the stuff reproduced:
   * The canonical WAVE format starts with the RIFF header:
   *
   * 0         4   ChunkID          Contains the letters "RIFF" in ASCII form
   *                                (0x52494646 big-endian form).
   * 4         4   ChunkSize        36 + SubChunk2Size, or more precisely:
   *                                4 + (8 + SubChunk1Size) + (8 + SubChunk2Size)
   *                                This is the size of the rest of the chunk
   *                                following this number.  This is the size of the
   *                                entire file in bytes minus 8 bytes for the
   *                                two fields not included in this count:
   *                                ChunkID and ChunkSize.
   * 8         4   Format           Contains the letters "WAVE"
   *                                (0x57415645 big-endian form).
   *
   * The "WAVE" format consists of two subchunks: "fmt " and "data":
   * The "fmt " subchunk describes the sound data's format:
   *
   * 12        4   Subchunk1ID      Contains the letters "fmt "
   *                                (0x666d7420 big-endian form).
   * 16        4   Subchunk1Size    16 for PCM.  This is the size of the
   *                                rest of the Subchunk which follows this number.
   * 20        2   AudioFormat      PCM = 1 (i.e. Linear quantization)
   *                                Values other than 1 indicate some
   *                                form of compression.
   * 22        2   NumChannels      Mono = 1, Stereo = 2, etc.
   * 24        4   SampleRate       8000, 44100, etc.
   * 28        4   ByteRate         == SampleRate * NumChannels * BitsPerSample/8
   * 32        2   BlockAlign       == NumChannels * BitsPerSample/8
   *                                The number of bytes for one sample including
   *                                all channels. I wonder what happens when
   *                                this number isn't an integer?
   * 34        2   BitsPerSample    8 bits = 8, 16 bits = 16, etc.
   *           2   ExtraParamSize   if PCM, then doesn't exist
   *           X   ExtraParams      space for extra parameters
   *
   * The "data" subchunk contains the size of the data and the actual sound:
   *
   * 36        4   Subchunk2ID      Contains the letters "data"
   *                                (0x64617461 big-endian form).
   * 40        4   Subchunk2Size    == NumSamples * NumChannels * BitsPerSample/8
   *                                This is the number of bytes in the data.
   *                                You can also think of this as the size
   *                                of the read of the subchunk following this
   *                                number.
   * 44        *   Data             The actual sound data.
   */

  if (stat->st_size < 44)
    return -1;

  if (strncmp(data, "RIFF", 4))
    return -1;

  /* we don't check the size, let's be relaxed */

  if (strncmp(data+8, "WAVE", 4))
    return -1;

  if (strncmp(data+12, "fmt ", 4))
    return -1;

  if (mem_to_long(data+16) != 16)
    return -2;

  if (mem_to_short(data+20) != 1)
    return -2;

  nb_channels = mem_to_short(data+22);
  if (nb_channels != 1 && nb_channels != 2)
    return -2;

  /* 44100 Hz mandatory */
  if ((int)mem_to_short(data+24) != 44100)
    return -2;

  /* then stuff we don't care */

  if (mem_to_short(data+34) != 16)
    return -2;

  if (strncmp(data+36, "data", 4))
    return -1;

  /* we don't check the size, we'll take stat.st_size instead (rounded as
   * necessary), so we handle partly corrupted files.
   */

  if (nb_channels == 1)
    return WAV_TYPE_MONO;

  return WAV_TYPE_STEREO_LEFT;
}

/* return 0 if error */
input *new_wav(char *file, int type)
{
  input tmp;
  input *ret;
  struct stat stat;
  int f;
  int file_type;

  memset(&tmp, 0, sizeof(input));

  if (type != WAV_TYPE_MONO && type != WAV_TYPE_STEREO_LEFT && type != WAV_TYPE_STEREO_RIGHT) {
    reset_error();
    add_error_char(file);
    add_error_char(":\nThe type ");
    add_error_int(type);
    add_error_char(" is not supported.\n");
    display_error();
    return 0;
  }

  f = open(file, O_RDONLY);
  if (f == -1) {
    int serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(":\n");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    return 0;
  }

  if (fstat(f, &stat) == -1) {
    int serr = errno;
    reset_error();
    add_error_char(file);
    add_error_char(":\n");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    close(f);
    return 0;
  }

  tmp.data = memmap(&stat, f, file);

  if (tmp.data == (void *)-1) {
    close(f);
    return 0;
  }

  close(f);

  file_type = analyze_wav(tmp.data, &stat);
  if (file_type == -1 || file_type == -2) {
    memunmap(tmp.data, stat.st_size);
    reset_error();
    add_error_char(file);
    add_error_char(": ");
    if (file_type == -1)
      add_error_char("not a wav file (or a corrupted one)\n");
    else
      add_error_char("unsupported wav file type (only 44100Hz 16b mono/stereo are handled)\n");
    display_error();
    return 0;
  }

  if (  (file_type == WAV_TYPE_MONO && type != WAV_TYPE_MONO)
     || (file_type == WAV_TYPE_STEREO_LEFT && (type != WAV_TYPE_STEREO_LEFT && type != WAV_TYPE_STEREO_RIGHT))) {
    memunmap(tmp.data, stat.st_size);
    reset_error();
    add_error_char(file);
    add_error_char(": the file type is not compatible with what was requested\n");
    add_error_char("(this means we were trying to load a stereo file and we have a mono one or the reverse)\n");
    display_error();
    return 0;
  }

  tmp.size = (stat.st_size-44) / 2;
  if (file_type == WAV_TYPE_STEREO_LEFT) /* file type can not be ..._RIGHT, only ..._LEFT to mean stereo */
    tmp.size /= 2;

  if (!tmp.size) {
    memunmap(tmp.data, stat.st_size);
    reset_error();
    add_error_char(file);
    add_error_char(": the file has an invalid length.\n");
    display_error();
    return 0;
  }

  tmp.file = strdup(file);	/* checked below, don't panic */
  tmp.file_size = stat.st_size;

  ret = malloc(sizeof(input));

  if (!tmp.file || !ret) {
    if (tmp.file) free(tmp.file);
    if (ret) free(ret);
    oom();
    display_error();
    memunmap(tmp.data, tmp.file_size);
    return 0;
  }

  *ret = tmp;

  ret->c.type = CHANNEL_TYPE_INPUT;
  ret->input_type = type;

  ret->open_stream = open_stream;
  ret->close_stream = close_stream;

  if (type == WAV_TYPE_MONO) {
    ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))mono_short_little_feed_buffer_short_mono;
    ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))mono_short_little_feed_buffer_float_mono;
    ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))mono_short_little_feed_buffer_short_stereo;
    ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))mono_short_little_feed_buffer_float_stereo;
    ret->float_stream = wav_mono_float_stream;
  }

  if (type == WAV_TYPE_STEREO_LEFT) {
    ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_left_short_little_feed_buffer_short_mono;
    ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_left_short_little_feed_buffer_float_mono;
    ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_left_short_little_feed_buffer_short_stereo;
    ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_left_short_little_feed_buffer_float_stereo;
    ret->float_stream = wav_stereo_left_float_stream;
  }

  if (type == WAV_TYPE_STEREO_RIGHT) {
    ret->c.feed_buffer_short_mono = (int (*)(void *, short *, int, int))stereo_right_short_little_feed_buffer_short_mono;
    ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))stereo_right_short_little_feed_buffer_float_mono;
    ret->c.feed_buffer_short_stereo = (int (*)(void *, short *, int, int))stereo_right_short_little_feed_buffer_short_stereo;
    ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))stereo_right_short_little_feed_buffer_float_stereo;
    ret->float_stream = wav_stereo_right_float_stream;
  }

  ret->c.first_point = (int (*)(void *))input_first_point;
  ret->c.last_point = (int (*)(void *))input_last_point;

  ret->c.reset = (void (*)(void *))input_reset;

  /* the first 44 bytes have to be "forgotten" */
  ret->data = (char *)ret->data + 44;

  return ret;
}

/* return 0 if yes,
 * -1 if not a wav
 * -2 if unsupported wav
 */
int is_wav(char *file)
{
  int f = open(file, O_RDONLY);
  struct stat stat;
  char buf[44];

  if (f == -1)
    return -1;

  if (fstat(f, &stat) == -1) {
    close(f);
    return -1;
  }

  if (read(f, buf, 44) != 44) {
    close(f);
    return -1;
  }

  close(f);

  return analyze_wav(buf, &stat);
}

void wav_stereo_create_header(char *b, int nb_samples, int stereo)
{
  int sub_chunk1_size;
  int sub_chunk2_size;
  int chunk_size;
  int bits_per_sample;
  int sample_rate;
  int num_channels;
  int audio_format;
  int byte_rate;
  int block_align;

  sub_chunk2_size = nb_samples * (stereo+1) * 2;
  sub_chunk1_size = 16;

  chunk_size = 4 + 8 + sub_chunk1_size + 8 + sub_chunk2_size;

  bits_per_sample = 16;
  sample_rate = 44100;
  num_channels = stereo+1;
  audio_format = 1;

  byte_rate = sample_rate * num_channels * (bits_per_sample / 8);

  block_align = num_channels * bits_per_sample / 8;

  memcpy(b, "RIFF", 4);
  long_to_mem(b+4, chunk_size);
  memcpy(b+8, "WAVE", 4);
  memcpy(b+12, "fmt ", 4);
  long_to_mem(b+16, sub_chunk1_size);
  short_to_mem(b+20, audio_format);
  short_to_mem(b+22, num_channels);
  long_to_mem(b+24, sample_rate);
  long_to_mem(b+28, byte_rate);
  short_to_mem(b+32, block_align);
  short_to_mem(b+34, bits_per_sample);
  memcpy(b+36, "data", 4);
  long_to_mem(b+40, sub_chunk2_size);
}
